using System;
using System.IO;
using System.Text;
using System.Runtime.InteropServices;

namespace CnBlogs.Helper.hBifTs
{
	#region Mapping File Enum Info
	public enum ProtectionTypes:int
	{
		PageReadOnly = 0x2,
		PageReadWrite = 0x4,
		PageWriteCopy = 0x8
	}

	public enum SectionTypes:int
	{
		SecCommit = 0x8000000,
		SecImage = 0x1000000,
		SecNoCache = 0x10000000,
		SecReserve = 0x4000000
	}

	public enum AccessTypes:int
	{
		Copy = 0x1,
		Read = 0x4,
		Write = 0x2,
		Full = 0xF001F
	}
	#endregion

	#region SharedMemory
	public class SharedMemory : IDisposable
	{
		#region DllImport
		[StructLayout(LayoutKind.Sequential)]
			private struct SECURITY_ATTRIBUTES
		{
			int nLength; 
			int lpSecurityDescriptor; 
			bool bInheritHandle; 
		}

		[DllImport("kernel32",CharSet = CharSet.Auto)]
		private static extern int CreateFileMapping( 
			int hFile,
			int lpSecurityAttributes, 
			int flProtect, 
			int dwMaximumSizeHigh, 
			int dwMaximumSizeLow, 
			[MarshalAs(UnmanagedType.LPTStr)] string lpName
			);

		[DllImport("kernel32",CharSet = CharSet.Auto)]
		private static extern int OpenFileMapping(
			int dwDesiredAccess,  
			bool bInheritHandle,
			[MarshalAs(UnmanagedType.LPTStr)] string lpName
			);

		[DllImport("kernel32",CharSet = CharSet.Auto)]
		private static extern bool UnmapViewOfFile(int lpBaseAddress);

		[DllImport("kernel32",CharSet = CharSet.Auto)]
		private static extern void CloseHandle(int Handle);

		[DllImport("kernel32",CharSet = CharSet.Auto)]
		private static extern int MapViewOfFile(
			int hFileMappingObject,
			int dwDesiredAccess,
			int dwFileOffsetHigh, 
			int dwFileOffsetLow, 
			int dwNumberOfBytesToMap
			);
		#endregion

		#region Private & Public Variable
		protected const int INVALID_HANDLE_VALUE = -1;
		protected int hMap = INVALID_HANDLE_VALUE;
		protected int lpAddr = 0;
		protected int size;
		protected int length;
		protected string name;

		public string Name
		{
			get{ return name;}
		}
		
		public int MapHandle
		{
			get{return hMap;}
		}
		
		public int Address
		{
			get{return lpAddr;}
		}

		public virtual int Size
		{
			get{ return size;}
		}

		public int Length
		{
			get{ return length;}
		}
		#endregion

		#region Constructor
		public SharedMemory(){}
		public SharedMemory(int Size,string Name) : this( Size,Name,ProtectionTypes.PageReadWrite,SectionTypes.SecCommit,AccessTypes.Full){}
		public SharedMemory(int Size, string Name, ProtectionTypes Protection, SectionTypes Section, AccessTypes Access	)
		{
			Create(Size, Name, Protection, Section, Access);
		}

		#endregion

		#region Common Function
		public virtual void Create( int Size,string Name)
		{
			Create( Size,Name,ProtectionTypes.PageReadWrite,SectionTypes.SecCommit,AccessTypes.Full);
		}

		public virtual void Create( int Size,string Name,int lpSecurityAttributes)
		{
			Create( Size,Name,lpSecurityAttributes,ProtectionTypes.PageReadWrite,SectionTypes.SecCommit,AccessTypes.Full);
		}

		public virtual void Create(	int Size, string Name,ProtectionTypes Protection, SectionTypes Section, AccessTypes Access)
		{
			Create(Size,Name,0,Protection,Section,Access);
		}

		public virtual void Create(int Size, string Name,int lpSecurityAttributes,ProtectionTypes Protection, SectionTypes Section, AccessTypes Access)
		{
			Close();
			size = Size;
			name = Name;
			hMap = CreateFileMapping(INVALID_HANDLE_VALUE, lpSecurityAttributes, (int)Protection|(int)Section, 0, Size, Name);
			if(hMap ==0)
			{
				throw new ApplicationException("Can't Create MappingFile " + Name);
			}
			lpAddr = MapViewOfFile(hMap, (int)Access, 0, 0, Size);
		}

		public virtual void Clear()
		{
			byte[] buffer = new byte[size];
			Marshal.Copy( buffer,0,new IntPtr(Address),size);
		}

		public virtual void Open(int Size, string Name){this.Open(Size, Name, 0, AccessTypes.Read);}
		public virtual void Open(int Size, string Name, AccessTypes Access){this.Open(Size, Name, 0, Access);}
		public virtual void Open(int Size, string Name, int Offset){this.Open(Size, Name, Offset, AccessTypes.Read);}
		public virtual void Open(int Size, string Name, int Offset, AccessTypes Access)
		{
			Close();
			name = Name;
			size = Size;
			hMap = OpenFileMapping((int)Access, false, Name);
			if(hMap == 0)
			{
				throw new ApplicationException("Can't Open MappingFile " + Name);
			}
			lpAddr = MapViewOfFile(hMap, (int)Access, 0, Offset, Size);
		}

		public virtual void CopyFrom(byte[] source,int startIndex,int length)
		{
			this.length = length;
			Clear();
			Marshal.Copy(source,startIndex,new IntPtr(Address),length);
		}

		public virtual void CopyFrom(byte[] source)
		{
			CopyFrom(source,0,source.Length);
		}

		public virtual void CopyFrom(string source)
		{
			byte[] temp = Encoding.Unicode.GetBytes(source);
			CopyFrom(temp,0,temp.Length);
		}

		public virtual void CopyTo(byte[] destination,int startIndex,int length)
		{
			Marshal.Copy( new IntPtr(Address),destination,startIndex,length);
		}

		public virtual byte[] CopyToBytes()
		{
			byte[] destination = new byte[Size];
			CopyTo(destination,0,Size);
			return destination;
		}

		public virtual byte[] CopyToBytes(int startIndex,int length)
		{
			byte[] destination = new byte[length];
			CopyTo(destination,startIndex,length);
			return destination;
		}

		public virtual string CopyToString()
		{
			return Encoding.Unicode.GetString( CopyToBytes() ).Replace("\0","").Trim();
		}

		public virtual string CopyToString(int startIndex,int length)
		{
			return Encoding.Unicode.GetString( CopyToBytes(startIndex,length) ).Replace("\0","").Trim();
		}

		public virtual void Close()
		{
			if( lpAddr != 0 )
			{
				UnmapViewOfFile(lpAddr);
				lpAddr = 0;
			}
			if( hMap != 0 && hMap != INVALID_HANDLE_VALUE )
			{
				CloseHandle(hMap);
				hMap = INVALID_HANDLE_VALUE;
			}
		}

		public virtual void Dispose()
		{
			Close();
		}
		#endregion
	}

	#endregion

	#region SharedMemoryEx
	class SharedMemoryEx : SharedMemory
	{

		#region Private Variable & Property
		private bool creator;

		private SharedMemory msgmem;
		private string msgmemname;
		private const int msgmemsize = 512;
		private const string CHANGESIZE = "ChangeSize";
		private const char SpliteChar	= ':';

		private int curlpSecuAttrs;
		private ProtectionTypes curProtection;
		private SectionTypes curSection;
		private AccessTypes curAccess;
		private int curOffset;


		public override int Size
		{
			get
			{
				CheckTo();
				return base.Size;
			}
		}

		#endregion

		#region Constructor & Dispose
		public SharedMemoryEx():base(){}

		public SharedMemoryEx(int Size,string Name):this(Size,Name,ProtectionTypes.PageReadWrite,SectionTypes.SecCommit,AccessTypes.Full){}

		public SharedMemoryEx(int Size, string Name, ProtectionTypes Protection, SectionTypes Section, AccessTypes Access):base(Size,Name,Protection,Section,Access){}

		public override void Dispose()
		{
			msgmem.Dispose();
			base.Dispose ();
		}
		#endregion

		#region Common Functions

		#region Open
		public override void Open(int Size, string Name, int Offset, AccessTypes Access)
		{
			creator = false;
			if( msgmem == null)
			{
				msgmem = new SharedMemory();
				msgmemname = GetMsgMemName(Name);
				msgmem.Open(msgmemsize,msgmemname,0,AccessTypes.Full);
			}

			curOffset = Offset;
			curAccess = Access;
			base.Open(Size,Name,Offset,Access);
		}

		public virtual void Open(string Name)
		{
			Open( Name, 0, AccessTypes.Read);
		}

		public virtual void Open(string Name, AccessTypes Access)
		{
			Open( Name, 0, Access);
		}

		public virtual void Open(string Name, int Offset)
		{
			Open(Name, Offset, AccessTypes.Read);
		}

		public virtual void Open(string Name,int Offset,AccessTypes Access)
		{
			Open(Name,Offset,Access,msgmemname);
		}

		public virtual void Open(string Name,string MsgMemName)
		{
			Open( Name, 0, AccessTypes.Read,MsgMemName);
		}

		public virtual void Open(string Name, AccessTypes Access,string MsgMemName)
		{
			Open( Name, 0, Access,MsgMemName);
		}

		public virtual void Open(string Name, int Offset,string MsgMemName)
		{
			Open(Name, Offset, AccessTypes.Read,MsgMemName);
		}

		public virtual void Open( string Name, int Offset, AccessTypes Access,string MsgMemName)
		{
			Open(0,Name,Offset,Access,MsgMemName);
		}

		public virtual void Open(int Size, string Name, int Offset, AccessTypes Access,string MsgMemName)
		{
			Open(Size,Name,Offset,Access);
		}
		#endregion

		#region Create
		public override void Create(int Size, string Name, int lpSecurityAttributes, ProtectionTypes Protection, SectionTypes Section, AccessTypes Access)
		{
			creator = true;
			if( msgmem == null)
			{
				msgmem = new SharedMemory();
				msgmemname = GetMsgMemName(Name);
				msgmem.Create(msgmemsize,msgmemname,ProtectionTypes.PageReadWrite,SectionTypes.SecCommit,AccessTypes.Full);
				SendChangeMsg(Size);
			}
			curlpSecuAttrs = lpSecurityAttributes;
			curProtection = Protection;
			curSection = Section;
			curAccess = Access;
			base.Create (Size, Name, lpSecurityAttributes, Protection, Section, Access);
		}
		
		public void Create(int Size, string Name, int lpSecurityAttributes, ProtectionTypes Protection, SectionTypes Section, AccessTypes Access,string MsgMemName)
		{
			msgmemname = MsgMemName;
			Create(Size,Name,lpSecurityAttributes,Protection,Section,Access);
		}

		#endregion

		#region CopyXXX 
		public override void CopyFrom(byte[] source, int startIndex, int length)
		{
			CheckFrom(length);
			base.CopyFrom (source, startIndex, length);
		}

		private void CheckFrom(int length)
		{
			if( length > Size)
			{
				if( !creator)
				{
					throw new ApplicationException("Can't Write so many bytes. Writalbe size " + Size + " .Want write size " + length);
				}
				//				byte[] msg = Encoding.Unicode.GetBytes( CHANGESIZE + SpliteChar);
				//				byte[] bysizes = BitConverter.GetBytes(length);
				//				byte[] temp = new byte[msg.Length + bysizes.Length];
				//				Array.Copy(msg,0,temp,0,msg.Length);
				//				Array.Copy(bysizes,0,temp,msg.Length,bysizes.Length);
				SendChangeMsg(length);
				Create(length,Name,curlpSecuAttrs,curProtection,curSection,curAccess);
			}
		}

		private void SendChangeMsg(int length)
		{
			msgmem.CopyFrom( CHANGESIZE + SpliteChar + length);
		}

		public override void CopyTo( byte[] destination, int startIndex, int length)
		{
			CheckTo();
			base.CopyTo (destination, startIndex, length);
		}

		private void CheckTo()
		{
			if( creator )
			{
				return;
			}
			int realsize = ReceiveChangeMsg();
			if( size < realsize)
			{
				Open( realsize,Name,curOffset,curAccess);
			}
		}

		private void ReceiveChangeMsg(out int length)
		{
			length = ReceiveChangeMsg();
		}

		private int ReceiveChangeMsg()
		{
			string[] msgs = msgmem.CopyToString().Split( SpliteChar);
			if( msgs.Length != 2 || msgs[0] != CHANGESIZE)
			{
				return 0;
			}
			return int.Parse( msgs[1]);
		}
		#endregion

		#endregion

		private static string GetMsgMemName(string Name)
		{
			byte[] temp = new byte[5];
			byte[] byname = Encoding.Unicode.GetBytes(Name);
			Array.Copy( byname,0,temp,0, (byname.Length >= temp.Length) ? temp.Length : byname.Length);
			for(int i=0;i< temp.Length; i++)
			{
				Name += temp[i];
			}
			return Name;
		}
	}
	#endregion
}
